@@ -26,7 +26,7 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: px.c,v 1.6 2001/11/20 18:54:07 momjian Exp $
+ * $PostgreSQL: pgsql/contrib/pgcrypto/px.c,v 1.9.4.1 2007/08/23 16:16:11 tgl Exp $
  */
 
 #include <postgres.h>
@@ -39,7 +39,7 @@ px_resolve_alias(const PX_Alias * list, const char *name)
 {
 	while (list->name)
 	{
-		if (!strcasecmp(list->alias, name))
+		if (pg_strcasecmp(list->alias, name) == 0)
 			return list->name;
 		list++;
 	}
@@ -185,6 +185,18 @@ combo_decrypt(PX_Combo * cx, const uint8 *data, unsigned dlen,
 
 	PX_Cipher  *c = cx->cipher;
 
+	/* decide whether zero-length input is allowed */
+	if (dlen == 0)
+	{
+		/* with padding, empty ciphertext is not allowed */
+		if (cx->padding)
+			return -1;
+		
+		/* without padding, report empty result */
+		*rlen = 0;
+		return 0;
+	}
+
 	bs = px_cipher_block_size(c);
 	if (bs > 1 && (dlen % bs) != 0)
 		goto block_error;
@@ -217,7 +229,7 @@ combo_decrypt(PX_Combo * cx, const uint8 *data, unsigned dlen,
 
 	/* error reporting should be done in pgcrypto.c */
 block_error:
-	elog(NOTICE, "Data not a multiple of block size");
+	elog(WARNING, "Data not a multiple of block size");
 	return -1;
 }
 
