@@ -4,11 +4,11 @@
  *	  Search code for postgres btrees.
  *
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/access/nbtree/nbtsearch.c,v 1.80 2003/08/08 21:41:27 momjian Exp $
+ *	  $PostgreSQL: pgsql/src/backend/access/nbtree/nbtsearch.c,v 1.90 2004/12/31 21:59:22 pgsql Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -17,6 +17,7 @@
 
 #include "access/genam.h"
 #include "access/nbtree.h"
+#include "utils/lsyscache.h"
 
 
 static Buffer _bt_walk_left(Relation rel, Buffer buf);
@@ -27,6 +28,10 @@ static bool _bt_endpoint(IndexScanDesc scan, ScanDirection dir);
  *	_bt_search() -- Search the tree for a particular scankey,
  *		or more precisely for the first leaf page it could be on.
  *
+ * When nextkey is false (the usual case), we are looking for the first
+ * item >= scankey.  When nextkey is true, we are looking for the first
+ * item strictly greater than scankey.
+ *
  * Return value is a stack of parent-page pointers.  *bufP is set to the
  * address of the leaf-page buffer, which is read-locked and pinned.
  * No locks are held on the parent pages, however!
@@ -37,7 +42,7 @@ static bool _bt_endpoint(IndexScanDesc scan, ScanDirection dir);
  * will result in *bufP being set to InvalidBuffer.
  */
 BTStack
-_bt_search(Relation rel, int keysz, ScanKey scankey,
+_bt_search(Relation rel, int keysz, ScanKey scankey, bool nextkey,
 		   Buffer *bufP, int access)
 {
 	BTStack		stack_in = NULL;
@@ -67,7 +72,7 @@ _bt_search(Relation rel, int keysz, ScanKey scankey,
 		 * its pointer in the parent (or metapage).  If it has, we may
 		 * need to move right to its new sibling.  Do that.
 		 */
-		*bufP = _bt_moveright(rel, *bufP, keysz, scankey, BT_READ);
+		*bufP = _bt_moveright(rel, *bufP, keysz, scankey, nextkey, BT_READ);
 
 		/* if this is a leaf page, we're done */
 		page = BufferGetPage(*bufP);
@@ -79,7 +84,7 @@ _bt_search(Relation rel, int keysz, ScanKey scankey,
 		 * Find the appropriate item on the internal page, and get the
 		 * child page that it points to.
 		 */
-		offnum = _bt_binsrch(rel, *bufP, keysz, scankey);
+		offnum = _bt_binsrch(rel, *bufP, keysz, scankey, nextkey);
 		itemid = PageGetItemId(page, offnum);
 		btitem = (BTItem) PageGetItem(page, itemid);
 		itup = &(btitem->bti_itup);
@@ -103,8 +108,7 @@ _bt_search(Relation rel, int keysz, ScanKey scankey,
 		new_stack->bts_parent = stack_in;
 
 		/* drop the read lock on the parent page, acquire one on the child */
-		_bt_relbuf(rel, *bufP);
-		*bufP = _bt_getbuf(rel, blkno, BT_READ);
+		*bufP = _bt_relandgetbuf(rel, *bufP, blkno, BT_READ);
 
 		/* okay, all set to move down a level */
 		stack_in = new_stack;
@@ -116,53 +120,65 @@ _bt_search(Relation rel, int keysz, ScanKey scankey,
 /*
  *	_bt_moveright() -- move right in the btree if necessary.
  *
- *		When we follow a pointer to reach a page, it is possible that
- *		the page has changed in the meanwhile.	If this happens, we're
- *		guaranteed that the page has "split right" -- that is, that any
- *		data that appeared on the page originally is either on the page
- *		or strictly to the right of it.
+ * When we follow a pointer to reach a page, it is possible that
+ * the page has changed in the meanwhile.  If this happens, we're
+ * guaranteed that the page has "split right" -- that is, that any
+ * data that appeared on the page originally is either on the page
+ * or strictly to the right of it.
  *
- *		This routine decides whether or not we need to move right in the
- *		tree by examining the high key entry on the page.  If that entry
- *		is strictly less than one we expect to be on the page, then our
- *		picture of the page is incorrect and we need to move right.
+ * When nextkey is false (the usual case), we are looking for the first
+ * item >= scankey.  When nextkey is true, we are looking for the first
+ * item strictly greater than scankey.
  *
- *		On entry, we have the buffer pinned and a lock of the proper type.
- *		If we move right, we release the buffer and lock and acquire the
- *		same on the right sibling.	Return value is the buffer we stop at.
+ * This routine decides whether or not we need to move right in the
+ * tree by examining the high key entry on the page.  If that entry
+ * is strictly less than the scankey, or <= the scankey in the nextkey=true
+ * case, then we followed the wrong link and we need to move right.
+ *
+ * On entry, we have the buffer pinned and a lock of the type specified by
+ * 'access'.  If we move right, we release the buffer and lock and acquire
+ * the same on the right sibling.  Return value is the buffer we stop at.
  */
 Buffer
 _bt_moveright(Relation rel,
 			  Buffer buf,
 			  int keysz,
 			  ScanKey scankey,
+			  bool nextkey,
 			  int access)
 {
 	Page		page;
 	BTPageOpaque opaque;
+	int32		cmpval;
 
 	page = BufferGetPage(buf);
 	opaque = (BTPageOpaque) PageGetSpecialPointer(page);
 
 	/*
-	 * If the scan key that brought us to this page is > the high key
-	 * stored on the page, then the page has split and we need to move
-	 * right.  (If the scan key is equal to the high key, we might or
-	 * might not need to move right; have to scan the page first anyway.)
-	 * It could even have split more than once, so scan as far as needed.
+	 * When nextkey = false (normal case): if the scan key that brought us
+	 * to this page is > the high key stored on the page, then the page
+	 * has split and we need to move right.  (If the scan key is equal to
+	 * the high key, we might or might not need to move right; have to
+	 * scan the page first anyway.)
+	 *
+	 * When nextkey = true: move right if the scan key is >= page's high key.
+	 *
+	 * The page could even have split more than once, so scan as far as
+	 * needed.
 	 *
 	 * We also have to move right if we followed a link that brought us to a
 	 * dead page.
 	 */
+	cmpval = nextkey ? 0 : 1;
+
 	while (!P_RIGHTMOST(opaque) &&
 		   (P_IGNORE(opaque) ||
-			_bt_compare(rel, keysz, scankey, page, P_HIKEY) > 0))
+			_bt_compare(rel, keysz, scankey, page, P_HIKEY) >= cmpval))
 	{
 		/* step right one page */
 		BlockNumber rblkno = opaque->btpo_next;
 
-		_bt_relbuf(rel, buf);
-		buf = _bt_getbuf(rel, rblkno, access);
+		buf = _bt_relandgetbuf(rel, buf, rblkno, access);
 		page = BufferGetPage(buf);
 		opaque = (BTPageOpaque) PageGetSpecialPointer(page);
 	}
@@ -177,21 +193,26 @@ _bt_moveright(Relation rel,
 /*
  *	_bt_binsrch() -- Do a binary search for a key on a particular page.
  *
+ * When nextkey is false (the usual case), we are looking for the first
+ * item >= scankey.  When nextkey is true, we are looking for the first
+ * item strictly greater than scankey.
+ *
  * The scankey we get has the compare function stored in the procedure
  * entry of each data struct.  We invoke this regproc to do the
  * comparison for every key in the scankey.
  *
  * On a leaf page, _bt_binsrch() returns the OffsetNumber of the first
- * key >= given scankey.  (NOTE: in particular, this means it is possible
- * to return a value 1 greater than the number of keys on the page,
- * if the scankey is > all keys on the page.)
+ * key >= given scankey, or > scankey if nextkey is true.  (NOTE: in
+ * particular, this means it is possible to return a value 1 greater than the
+ * number of keys on the page, if the scankey is > all keys on the page.)
  *
  * On an internal (non-leaf) page, _bt_binsrch() returns the OffsetNumber
- * of the last key < given scankey.  (Since _bt_compare treats the first
- * data key of such a page as minus infinity, there will be at least one
- * key < scankey, so the result always points at one of the keys on the
- * page.)  This key indicates the right place to descend to be sure we
- * find all leaf keys >= given scankey.
+ * of the last key < given scankey, or last key <= given scankey if nextkey
+ * is true.  (Since _bt_compare treats the first data key of such a page as
+ * minus infinity, there will be at least one key < scankey, so the result
+ * always points at one of the keys on the page.)  This key indicates the
+ * right place to descend to be sure we find all leaf keys >= given scankey
+ * (or leaf keys > given scankey when nextkey is true).
  *
  * This procedure is not responsible for walking right, it just examines
  * the given page.	_bt_binsrch() has no lock or refcount side effects
@@ -201,14 +222,16 @@ OffsetNumber
 _bt_binsrch(Relation rel,
 			Buffer buf,
 			int keysz,
-			ScanKey scankey)
+			ScanKey scankey,
+			bool nextkey)
 {
 	TupleDesc	itupdesc;
 	Page		page;
 	BTPageOpaque opaque;
 	OffsetNumber low,
 				high;
-	int32		result;
+	int32		result,
+				cmpval;
 
 	itupdesc = RelationGetDescr(rel);
 	page = BufferGetPage(buf);
@@ -228,12 +251,21 @@ _bt_binsrch(Relation rel,
 		return low;
 
 	/*
-	 * Binary search to find the first key on the page >= scan key. Loop
-	 * invariant: all slots before 'low' are < scan key, all slots at or
-	 * after 'high' are >= scan key.  We can fall out when high == low.
+	 * Binary search to find the first key on the page >= scan key, or
+	 * first key > scankey when nextkey is true.
+	 *
+	 * For nextkey=false (cmpval=1), the loop invariant is: all slots before
+	 * 'low' are < scan key, all slots at or after 'high' are >= scan key.
+	 *
+	 * For nextkey=true (cmpval=0), the loop invariant is: all slots before
+	 * 'low' are <= scan key, all slots at or after 'high' are > scan key.
+	 *
+	 * We can fall out when high == low.
 	 */
 	high++;						/* establish the loop invariant for high */
 
+	cmpval = nextkey ? 0 : 1;	/* select comparison value */
+
 	while (high > low)
 	{
 		OffsetNumber mid = low + ((high - low) / 2);
@@ -242,7 +274,7 @@ _bt_binsrch(Relation rel,
 
 		result = _bt_compare(rel, keysz, scankey, page, mid);
 
-		if (result > 0)
+		if (result >= cmpval)
 			low = mid + 1;
 		else
 			high = mid;
@@ -252,15 +284,15 @@ _bt_binsrch(Relation rel,
 	 * At this point we have high == low, but be careful: they could point
 	 * past the last slot on the page.
 	 *
-	 * On a leaf page, we always return the first key >= scan key (which
-	 * could be the last slot + 1).
+	 * On a leaf page, we always return the first key >= scan key (resp. >
+	 * scan key), which could be the last slot + 1.
 	 */
 	if (P_ISLEAF(opaque))
 		return low;
 
 	/*
-	 * On a non-leaf page, return the last key < scan key. There must be
-	 * one if _bt_compare() is playing by the rules.
+	 * On a non-leaf page, return the last key < scan key (resp. <= scan
+	 * key). There must be one if _bt_compare() is playing by the rules.
 	 */
 	Assert(low > P_FIRSTDATAKEY(opaque));
 
@@ -325,17 +357,16 @@ _bt_compare(Relation rel,
 	 * (see _bt_first).
 	 */
 
-	for (i = 0; i < keysz; i++)
+	for (i = 1; i <= keysz; i++)
 	{
-		ScanKey		entry = &scankey[i];
 		Datum		datum;
 		bool		isNull;
 		int32		result;
 
-		datum = index_getattr(itup, entry->sk_attno, itupdesc, &isNull);
+		datum = index_getattr(itup, scankey->sk_attno, itupdesc, &isNull);
 
 		/* see comments about NULLs handling in btbuild */
-		if (entry->sk_flags & SK_ISNULL)		/* key is NULL */
+		if (scankey->sk_flags & SK_ISNULL)		/* key is NULL */
 		{
 			if (isNull)
 				result = 0;		/* NULL "=" NULL */
@@ -348,14 +379,28 @@ _bt_compare(Relation rel,
 		}
 		else
 		{
-			result = DatumGetInt32(FunctionCall2(&entry->sk_func,
-												 entry->sk_argument,
-												 datum));
+			/*
+			 * The sk_func needs to be passed the index value as left arg
+			 * and the sk_argument as right arg (they might be of
+			 * different types).  Since it is convenient for callers to
+			 * think of _bt_compare as comparing the scankey to the index
+			 * item, we have to flip the sign of the comparison result.
+			 *
+			 * Note: curious-looking coding is to avoid overflow if
+			 * comparison function returns INT_MIN.  There is no risk of
+			 * overflow for positive results.
+			 */
+			result = DatumGetInt32(FunctionCall2(&scankey->sk_func,
+												 datum,
+												 scankey->sk_argument));
+			result = (result < 0) ? 1 : -result;
 		}
 
 		/* if the keys are unequal, return the difference */
 		if (result != 0)
 			return result;
+
+		scankey++;
 	}
 
 	/* if we get here, the keys are equal */
@@ -447,136 +492,316 @@ _bt_first(IndexScanDesc scan, ScanDirection dir)
 	BlockNumber blkno;
 	StrategyNumber strat;
 	bool		res;
-	int32		result;
-	bool		scanFromEnd;
+	bool		nextkey;
+	bool		goback;
 	bool		continuescan;
-	ScanKey		scankeys = NULL;
+	ScanKey		scankeys;
+	ScanKey    *startKeys = NULL;
 	int			keysCount = 0;
-	int		   *nKeyIs = NULL;
-	int			i,
-				j;
+	int			i;
 	StrategyNumber strat_total;
 
 	/*
-	 * Order the scan keys in our canonical fashion and eliminate any
-	 * redundant keys.
+	 * Examine the scan keys and eliminate any redundant keys; also
+	 * discover how many keys must be matched to continue the scan.
 	 */
-	_bt_orderkeys(scan);
+	_bt_preprocess_keys(scan);
 
 	/*
-	 * Quit now if _bt_orderkeys() discovered that the scan keys can never
-	 * be satisfied (eg, x == 1 AND x > 2).
+	 * Quit now if _bt_preprocess_keys() discovered that the scan keys can
+	 * never be satisfied (eg, x == 1 AND x > 2).
 	 */
 	if (!so->qual_ok)
 		return false;
 
-	/*
+	/*----------
 	 * Examine the scan keys to discover where we need to start the scan.
+	 *
+	 * We want to identify the keys that can be used as starting boundaries;
+	 * these are =, >, or >= keys for a forward scan or =, <, <= keys for
+	 * a backwards scan.  We can use keys for multiple attributes so long as
+	 * the prior attributes had only =, >= (resp. =, <=) keys.	Once we accept
+	 * a > or < boundary or find an attribute with no boundary (which can be
+	 * thought of as the same as "> -infinity"), we can't use keys for any
+	 * attributes to its right, because it would break our simplistic notion
+	 * of what initial positioning strategy to use.
+	 *
+	 * When the scan keys include non-default operators, _bt_preprocess_keys
+	 * may not be able to eliminate redundant keys; in such cases we will
+	 * arbitrarily pick a usable one for each attribute.  This is correct
+	 * but possibly not optimal behavior.  (For example, with keys like
+	 * "x >= 4 AND x >= 5" we would elect to scan starting at x=4 when
+	 * x=5 would be more efficient.)  Since the situation only arises in
+	 * hokily-worded queries, live with it.
+	 *
+	 * When both equality and inequality keys appear for a single attribute
+	 * (again, only possible when non-default operators appear), we *must*
+	 * select one of the equality keys for the starting point, because
+	 * _bt_checkkeys() will stop the scan as soon as an equality qual fails.
+	 * For example, if we have keys like "x >= 4 AND x = 10" and we elect to
+	 * start at x=4, we will fail and stop before reaching x=10.  If multiple
+	 * equality quals survive preprocessing, however, it doesn't matter which
+	 * one we use --- by definition, they are either redundant or
+	 * contradictory.
+	 *----------
 	 */
-	scanFromEnd = false;
 	strat_total = BTEqualStrategyNumber;
 	if (so->numberOfKeys > 0)
 	{
-		nKeyIs = (int *) palloc(so->numberOfKeys * sizeof(int));
-		for (i = 0; i < so->numberOfKeys; i++)
+		AttrNumber	curattr;
+		ScanKey		chosen;
+		ScanKey		cur;
+
+		startKeys = (ScanKey *) palloc(so->numberOfKeys * sizeof(ScanKey));
+
+		/*
+		 * chosen is the so-far-chosen key for the current attribute, if
+		 * any. We don't cast the decision in stone until we reach keys
+		 * for the next attribute.
+		 */
+		curattr = 1;
+		chosen = NULL;
+
+		/*
+		 * Loop iterates from 0 to numberOfKeys inclusive; we use the last
+		 * pass to handle after-last-key processing.  Actual exit from the
+		 * loop is at one of the "break" statements below.
+		 */
+		for (cur = so->keyData, i = 0;; cur++, i++)
 		{
-			AttrNumber	attno = so->keyData[i].sk_attno;
-
-			/* ignore keys for already-determined attrs */
-			if (attno <= keysCount)
-				continue;
-			/* if we didn't find a boundary for the preceding attr, quit */
-			if (attno > keysCount + 1)
-				break;
-			strat = _bt_getstrat(rel, attno,
-								 so->keyData[i].sk_procedure);
-
-			/*
-			 * Can we use this key as a starting boundary for this attr?
-			 *
-			 * We can use multiple keys if they look like, say, = >= = but we
-			 * have to stop after accepting a > or < boundary.
-			 */
-			if (strat == strat_total ||
-				strat == BTEqualStrategyNumber)
-				nKeyIs[keysCount++] = i;
-			else if (ScanDirectionIsBackward(dir) &&
-					 (strat == BTLessStrategyNumber ||
-					  strat == BTLessEqualStrategyNumber))
+			if (i >= so->numberOfKeys || cur->sk_attno != curattr)
 			{
-				nKeyIs[keysCount++] = i;
-				strat_total = strat;
-				if (strat == BTLessStrategyNumber)
+				/*
+				 * Done looking at keys for curattr.  If we didn't find a
+				 * usable boundary key, quit; else save the boundary key
+				 * pointer in startKeys.
+				 */
+				if (chosen == NULL)
 					break;
+				startKeys[keysCount++] = chosen;
+
+				/*
+				 * Adjust strat_total, and quit if we have stored a > or <
+				 * key.
+				 */
+				strat = chosen->sk_strategy;
+				if (strat != BTEqualStrategyNumber)
+				{
+					strat_total = strat;
+					if (strat == BTGreaterStrategyNumber ||
+						strat == BTLessStrategyNumber)
+						break;
+				}
+
+				/*
+				 * Done if that was the last attribute.
+				 */
+				if (i >= so->numberOfKeys)
+					break;
+
+				/*
+				 * Reset for next attr, which should be in sequence.
+				 */
+				Assert(cur->sk_attno == curattr + 1);
+				curattr = cur->sk_attno;
+				chosen = NULL;
 			}
-			else if (ScanDirectionIsForward(dir) &&
-					 (strat == BTGreaterStrategyNumber ||
-					  strat == BTGreaterEqualStrategyNumber))
+
+			/* Can we use this key as a starting boundary for this attr? */
+			switch (cur->sk_strategy)
 			{
-				nKeyIs[keysCount++] = i;
-				strat_total = strat;
-				if (strat == BTGreaterStrategyNumber)
+				case BTLessStrategyNumber:
+				case BTLessEqualStrategyNumber:
+					if (chosen == NULL && ScanDirectionIsBackward(dir))
+						chosen = cur;
+					break;
+				case BTEqualStrategyNumber:
+					/* override any non-equality choice */
+					chosen = cur;
+					break;
+				case BTGreaterEqualStrategyNumber:
+				case BTGreaterStrategyNumber:
+					if (chosen == NULL && ScanDirectionIsForward(dir))
+						chosen = cur;
 					break;
 			}
 		}
-		if (keysCount == 0)
-			scanFromEnd = true;
 	}
-	else
-		scanFromEnd = true;
 
-	/* if we just need to walk down one edge of the tree, do that */
-	if (scanFromEnd)
+	/*
+	 * If we found no usable boundary keys, we have to start from one end
+	 * of the tree.  Walk down that edge to the first or last key, and
+	 * scan from there.
+	 */
+	if (keysCount == 0)
 	{
-		if (nKeyIs)
-			pfree(nKeyIs);
+		if (startKeys)
+			pfree(startKeys);
 		return _bt_endpoint(scan, dir);
 	}
 
 	/*
 	 * We want to start the scan somewhere within the index.  Set up a
-	 * scankey we can use to search for the correct starting point.
+	 * 3-way-comparison scankey we can use to search for the boundary
+	 * point we identified above.
 	 */
 	scankeys = (ScanKey) palloc(keysCount * sizeof(ScanKeyData));
 	for (i = 0; i < keysCount; i++)
 	{
-		FmgrInfo   *procinfo;
-
-		j = nKeyIs[i];
+		ScanKey		cur = startKeys[i];
 
 		/*
-		 * _bt_orderkeys disallows it, but it's place to add some code
-		 * later
+		 * _bt_preprocess_keys disallows it, but it's place to add some
+		 * code later
 		 */
-		if (so->keyData[j].sk_flags & SK_ISNULL)
+		if (cur->sk_flags & SK_ISNULL)
 		{
-			pfree(nKeyIs);
+			pfree(startKeys);
 			pfree(scankeys);
 			elog(ERROR, "btree doesn't support is(not)null, yet");
 			return false;
 		}
-		procinfo = index_getprocinfo(rel, i + 1, BTORDER_PROC);
-		ScanKeyEntryInitializeWithInfo(scankeys + i,
-									   so->keyData[j].sk_flags,
-									   i + 1,
-									   procinfo,
-									   CurrentMemoryContext,
-									   so->keyData[j].sk_argument);
-	}
-	if (nKeyIs)
-		pfree(nKeyIs);
 
-	current = &(scan->currentItemData);
+		/*
+		 * If scankey operator is of default subtype, we can use the
+		 * cached comparison procedure; otherwise gotta look it up in the
+		 * catalogs.
+		 */
+		if (cur->sk_subtype == InvalidOid)
+		{
+			FmgrInfo   *procinfo;
+
+			procinfo = index_getprocinfo(rel, i + 1, BTORDER_PROC);
+			ScanKeyEntryInitializeWithInfo(scankeys + i,
+										   cur->sk_flags,
+										   i + 1,
+										   InvalidStrategy,
+										   InvalidOid,
+										   procinfo,
+										   cur->sk_argument);
+		}
+		else
+		{
+			RegProcedure cmp_proc;
+
+			cmp_proc = get_opclass_proc(rel->rd_index->indclass[i],
+										cur->sk_subtype,
+										BTORDER_PROC);
+			ScanKeyEntryInitialize(scankeys + i,
+								   cur->sk_flags,
+								   i + 1,
+								   InvalidStrategy,
+								   cur->sk_subtype,
+								   cmp_proc,
+								   cur->sk_argument);
+		}
+	}
+
+	pfree(startKeys);
+
+	/*
+	 * Examine the selected initial-positioning strategy to determine
+	 * exactly where we need to start the scan, and set flag variables to
+	 * control the code below.
+	 *
+	 * If nextkey = false, _bt_search and _bt_binsrch will locate the first
+	 * item >= scan key.  If nextkey = true, they will locate the first
+	 * item > scan key.
+	 *
+	 * If goback = true, we will then step back one item, while if goback =
+	 * false, we will start the scan on the located item.
+	 *
+	 * it's yet other place to add some code later for is(not)null ...
+	 */
+	switch (strat_total)
+	{
+		case BTLessStrategyNumber:
+
+			/*
+			 * Find first item >= scankey, then back up one to arrive at
+			 * last item < scankey.  (Note: this positioning strategy is
+			 * only used for a backward scan, so that is always the
+			 * correct starting position.)
+			 */
+			nextkey = false;
+			goback = true;
+			break;
+
+		case BTLessEqualStrategyNumber:
+
+			/*
+			 * Find first item > scankey, then back up one to arrive at
+			 * last item <= scankey.  (Note: this positioning strategy is
+			 * only used for a backward scan, so that is always the
+			 * correct starting position.)
+			 */
+			nextkey = true;
+			goback = true;
+			break;
+
+		case BTEqualStrategyNumber:
+
+			/*
+			 * If a backward scan was specified, need to start with last
+			 * equal item not first one.
+			 */
+			if (ScanDirectionIsBackward(dir))
+			{
+				/*
+				 * This is the same as the <= strategy.  We will check at
+				 * the end whether the found item is actually =.
+				 */
+				nextkey = true;
+				goback = true;
+			}
+			else
+			{
+				/*
+				 * This is the same as the >= strategy.  We will check at
+				 * the end whether the found item is actually =.
+				 */
+				nextkey = false;
+				goback = false;
+			}
+			break;
+
+		case BTGreaterEqualStrategyNumber:
+
+			/*
+			 * Find first item >= scankey.	(This is only used for forward
+			 * scans.)
+			 */
+			nextkey = false;
+			goback = false;
+			break;
+
+		case BTGreaterStrategyNumber:
+
+			/*
+			 * Find first item > scankey.  (This is only used for forward
+			 * scans.)
+			 */
+			nextkey = true;
+			goback = false;
+			break;
+
+		default:
+			/* can't get here, but keep compiler quiet */
+			elog(ERROR, "unrecognized strat_total: %d", (int) strat_total);
+			return false;
+	}
 
 	/*
 	 * Use the manufactured scan key to descend the tree and position
 	 * ourselves on the target leaf page.
 	 */
-	stack = _bt_search(rel, keysCount, scankeys, &buf, BT_READ);
+	stack = _bt_search(rel, keysCount, scankeys, nextkey, &buf, BT_READ);
 
 	/* don't need to keep the stack around... */
 	_bt_freestack(stack);
 
+	current = &(scan->currentItemData);
+
 	if (!BufferIsValid(buf))
 	{
 		/* Only get here if index is completely empty */
@@ -588,167 +813,51 @@ _bt_first(IndexScanDesc scan, ScanDirection dir)
 
 	/* remember which buffer we have pinned */
 	so->btso_curbuf = buf;
-	blkno = BufferGetBlockNumber(buf);
-	page = BufferGetPage(buf);
 
 	/* position to the precise item on the page */
-	offnum = _bt_binsrch(rel, buf, keysCount, scankeys);
+	offnum = _bt_binsrch(rel, buf, keysCount, scankeys, nextkey);
 
+	page = BufferGetPage(buf);
+	blkno = BufferGetBlockNumber(buf);
 	ItemPointerSet(current, blkno, offnum);
 
+	/* done with manufactured scankey, now */
+	pfree(scankeys);
+
 	/*
-	 * At this point we are positioned at the first item >= scan key, or
+	 * If nextkey = false, we are positioned at the first item >= scan
+	 * key, or possibly at the end of a page on which all the existing
+	 * items are less than the scan key and we know that everything on
+	 * later pages is greater than or equal to scan key.
+	 *
+	 * If nextkey = true, we are positioned at the first item > scan key, or
 	 * possibly at the end of a page on which all the existing items are
-	 * less than the scan key and we know that everything on later pages
-	 * is greater than or equal to scan key.
+	 * less than or equal to the scan key and we know that everything on
+	 * later pages is greater than scan key.
 	 *
-	 * We could step forward in the latter case, but that'd be a waste of
-	 * time if we want to scan backwards.  So, it's now time to examine
-	 * the scan strategy to find the exact place to start the scan.
+	 * The actually desired starting point is either this item or the prior
+	 * one, or in the end-of-page case it's the first item on the next
+	 * page or the last item on this page.	We apply _bt_step if needed to
+	 * get to the right place.
 	 *
-	 * Note: if _bt_step fails (meaning we fell off the end of the index in
-	 * one direction or the other), we either return false (no matches) or
-	 * call _bt_endpoint() to set up a scan starting at that index
-	 * endpoint, as appropriate for the desired scan type.
-	 *
-	 * it's yet other place to add some code later for is(not)null ...
+	 * If _bt_step fails (meaning we fell off the end of the index in one
+	 * direction or the other), then there are no matches so we just
+	 * return false.
 	 */
-
-	switch (strat_total)
+	if (goback)
 	{
-		case BTLessStrategyNumber:
-
-			/*
-			 * Back up one to arrive at last item < scankey
-			 */
-			if (!_bt_step(scan, &buf, BackwardScanDirection))
-			{
-				pfree(scankeys);
+		/* _bt_step will do the right thing if we are at end-of-page */
+		if (!_bt_step(scan, &buf, BackwardScanDirection))
+			return false;
+	}
+	else
+	{
+		/* If we're at end-of-page, must step forward to next page */
+		if (offnum > PageGetMaxOffsetNumber(page))
+		{
+			if (!_bt_step(scan, &buf, ForwardScanDirection))
 				return false;
-			}
-			break;
-
-		case BTLessEqualStrategyNumber:
-
-			/*
-			 * We need to find the last item <= scankey, so step forward
-			 * till we find one > scankey, then step back one.
-			 */
-			if (offnum > PageGetMaxOffsetNumber(page))
-			{
-				if (!_bt_step(scan, &buf, ForwardScanDirection))
-				{
-					pfree(scankeys);
-					return _bt_endpoint(scan, dir);
-				}
-			}
-			for (;;)
-			{
-				offnum = ItemPointerGetOffsetNumber(current);
-				page = BufferGetPage(buf);
-				result = _bt_compare(rel, keysCount, scankeys, page, offnum);
-				if (result < 0)
-					break;
-				if (!_bt_step(scan, &buf, ForwardScanDirection))
-				{
-					pfree(scankeys);
-					return _bt_endpoint(scan, dir);
-				}
-			}
-			if (!_bt_step(scan, &buf, BackwardScanDirection))
-			{
-				pfree(scankeys);
-				return false;
-			}
-			break;
-
-		case BTEqualStrategyNumber:
-
-			/*
-			 * Make sure we are on the first equal item; might have to
-			 * step forward if currently at end of page.
-			 */
-			if (offnum > PageGetMaxOffsetNumber(page))
-			{
-				if (!_bt_step(scan, &buf, ForwardScanDirection))
-				{
-					pfree(scankeys);
-					return false;
-				}
-				offnum = ItemPointerGetOffsetNumber(current);
-				page = BufferGetPage(buf);
-			}
-			result = _bt_compare(rel, keysCount, scankeys, page, offnum);
-			if (result != 0)
-				goto nomatches; /* no equal items! */
-
-			/*
-			 * If a backward scan was specified, need to start with last
-			 * equal item not first one.
-			 */
-			if (ScanDirectionIsBackward(dir))
-			{
-				do
-				{
-					if (!_bt_step(scan, &buf, ForwardScanDirection))
-					{
-						pfree(scankeys);
-						return _bt_endpoint(scan, dir);
-					}
-					offnum = ItemPointerGetOffsetNumber(current);
-					page = BufferGetPage(buf);
-					result = _bt_compare(rel, keysCount, scankeys, page, offnum);
-				} while (result == 0);
-				if (!_bt_step(scan, &buf, BackwardScanDirection))
-					elog(ERROR, "equal items disappeared?");
-			}
-			break;
-
-		case BTGreaterEqualStrategyNumber:
-
-			/*
-			 * We want the first item >= scankey, which is where we are...
-			 * unless we're not anywhere at all...
-			 */
-			if (offnum > PageGetMaxOffsetNumber(page))
-			{
-				if (!_bt_step(scan, &buf, ForwardScanDirection))
-				{
-					pfree(scankeys);
-					return false;
-				}
-			}
-			break;
-
-		case BTGreaterStrategyNumber:
-
-			/*
-			 * We want the first item > scankey, so make sure we are on an
-			 * item and then step over any equal items.
-			 */
-			if (offnum > PageGetMaxOffsetNumber(page))
-			{
-				if (!_bt_step(scan, &buf, ForwardScanDirection))
-				{
-					pfree(scankeys);
-					return false;
-				}
-				offnum = ItemPointerGetOffsetNumber(current);
-				page = BufferGetPage(buf);
-			}
-			result = _bt_compare(rel, keysCount, scankeys, page, offnum);
-			while (result == 0)
-			{
-				if (!_bt_step(scan, &buf, ForwardScanDirection))
-				{
-					pfree(scankeys);
-					return false;
-				}
-				offnum = ItemPointerGetOffsetNumber(current);
-				page = BufferGetPage(buf);
-				result = _bt_compare(rel, keysCount, scankeys, page, offnum);
-			}
-			break;
+		}
 	}
 
 	/* okay, current item pointer for the scan is right */
@@ -772,15 +881,12 @@ _bt_first(IndexScanDesc scan, ScanDirection dir)
 	else
 	{
 		/* no tuples in the index match this scan key */
-nomatches:
 		ItemPointerSetInvalid(current);
 		so->btso_curbuf = InvalidBuffer;
 		_bt_relbuf(rel, buf);
 		res = false;
 	}
 
-	pfree(scankeys);
-
 	return res;
 }
 
@@ -836,8 +942,7 @@ _bt_step(IndexScanDesc scan, Buffer *bufP, ScanDirection dir)
 				}
 				/* step right one page */
 				blkno = opaque->btpo_next;
-				_bt_relbuf(rel, *bufP);
-				*bufP = _bt_getbuf(rel, blkno, BT_READ);
+				*bufP = _bt_relandgetbuf(rel, *bufP, blkno, BT_READ);
 				page = BufferGetPage(*bufP);
 				opaque = (BTPageOpaque) PageGetSpecialPointer(page);
 				if (!P_IGNORE(opaque))
@@ -852,8 +957,8 @@ _bt_step(IndexScanDesc scan, Buffer *bufP, ScanDirection dir)
 		}
 	}
 	else
-/* backwards scan */
 	{
+		/* backwards scan */
 		if (offnum > P_FIRSTDATAKEY(opaque))
 			offnum = OffsetNumberPrev(offnum);
 		else
@@ -944,8 +1049,7 @@ _bt_walk_left(Relation rel, Buffer buf)
 		obknum = BufferGetBlockNumber(buf);
 		/* step left */
 		blkno = lblkno = opaque->btpo_prev;
-		_bt_relbuf(rel, buf);
-		buf = _bt_getbuf(rel, blkno, BT_READ);
+		buf = _bt_relandgetbuf(rel, buf, blkno, BT_READ);
 		page = BufferGetPage(buf);
 		opaque = (BTPageOpaque) PageGetSpecialPointer(page);
 
@@ -972,15 +1076,13 @@ _bt_walk_left(Relation rel, Buffer buf)
 			if (P_RIGHTMOST(opaque) || ++tries > 4)
 				break;
 			blkno = opaque->btpo_next;
-			_bt_relbuf(rel, buf);
-			buf = _bt_getbuf(rel, blkno, BT_READ);
+			buf = _bt_relandgetbuf(rel, buf, blkno, BT_READ);
 			page = BufferGetPage(buf);
 			opaque = (BTPageOpaque) PageGetSpecialPointer(page);
 		}
 
 		/* Return to the original page to see what's up */
-		_bt_relbuf(rel, buf);
-		buf = _bt_getbuf(rel, obknum, BT_READ);
+		buf = _bt_relandgetbuf(rel, buf, obknum, BT_READ);
 		page = BufferGetPage(buf);
 		opaque = (BTPageOpaque) PageGetSpecialPointer(page);
 		if (P_ISDELETED(opaque))
@@ -997,8 +1099,7 @@ _bt_walk_left(Relation rel, Buffer buf)
 					elog(ERROR, "fell off the end of \"%s\"",
 						 RelationGetRelationName(rel));
 				blkno = opaque->btpo_next;
-				_bt_relbuf(rel, buf);
-				buf = _bt_getbuf(rel, blkno, BT_READ);
+				buf = _bt_relandgetbuf(rel, buf, blkno, BT_READ);
 				page = BufferGetPage(buf);
 				opaque = (BTPageOpaque) PageGetSpecialPointer(page);
 				if (!P_ISDELETED(opaque))
@@ -1080,8 +1181,7 @@ _bt_get_endpoint(Relation rel, uint32 level, bool rightmost)
 			if (blkno == P_NONE)
 				elog(ERROR, "fell off the end of \"%s\"",
 					 RelationGetRelationName(rel));
-			_bt_relbuf(rel, buf);
-			buf = _bt_getbuf(rel, blkno, BT_READ);
+			buf = _bt_relandgetbuf(rel, buf, blkno, BT_READ);
 			page = BufferGetPage(buf);
 			opaque = (BTPageOpaque) PageGetSpecialPointer(page);
 		}
@@ -1102,8 +1202,7 @@ _bt_get_endpoint(Relation rel, uint32 level, bool rightmost)
 		itup = &(btitem->bti_itup);
 		blkno = ItemPointerGetBlockNumber(&(itup->t_tid));
 
-		_bt_relbuf(rel, buf);
-		buf = _bt_getbuf(rel, blkno, BT_READ);
+		buf = _bt_relandgetbuf(rel, buf, blkno, BT_READ);
 		page = BufferGetPage(buf);
 		opaque = (BTPageOpaque) PageGetSpecialPointer(page);
 	}
@@ -1112,7 +1211,8 @@ _bt_get_endpoint(Relation rel, uint32 level, bool rightmost)
 }
 
 /*
- *	_bt_endpoint() -- Find the first or last key in the index.
+ *	_bt_endpoint() -- Find the first or last key in the index, and scan
+ * from there to the first key satisfying all the quals.
  *
  * This is used by _bt_first() to set up a scan when we've determined
  * that the scan must start at the beginning or end of the index (for
@@ -1202,7 +1302,10 @@ _bt_endpoint(IndexScanDesc scan, ScanDirection dir)
 	btitem = (BTItem) PageGetItem(page, PageGetItemId(page, start));
 	itup = &(btitem->bti_itup);
 
-	/* see if we picked a winner */
+	/*
+	 * Okay, we are on the first or last tuple.  Does it pass all the
+	 * quals?
+	 */
 	if (_bt_checkkeys(scan, itup, dir, &continuescan))
 	{
 		/* yes, return it */
@@ -1211,7 +1314,7 @@ _bt_endpoint(IndexScanDesc scan, ScanDirection dir)
 	}
 	else if (continuescan)
 	{
-		/* no, but there might be another one that is */
+		/* no, but there might be another one that does */
 		res = _bt_next(scan, dir);
 	}
 	else
