@@ -3,23 +3,25 @@
  * lmgr.c
  *	  POSTGRES lock manager code
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/storage/lmgr/lmgr.c,v 1.60 2003/09/04 22:06:27 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/storage/lmgr/lmgr.c,v 1.71 2004/12/31 22:01:05 pgsql Exp $
  *
  *-------------------------------------------------------------------------
  */
 
 #include "postgres.h"
 
+#include "access/subtrans.h"
 #include "access/transam.h"
 #include "access/xact.h"
 #include "catalog/catalog.h"
 #include "miscadmin.h"
 #include "storage/lmgr.h"
+#include "storage/sinval.h"
 #include "utils/inval.h"
 
 
@@ -65,26 +67,31 @@ static LOCKMASK LockConflicts[] = {
 
 };
 
-LOCKMETHOD	LockTableId = (LOCKMETHOD) NULL;
-LOCKMETHOD	LongTermTableId = (LOCKMETHOD) NULL;
+static LOCKMETHODID LockTableId = INVALID_LOCKMETHOD;
 
 /*
  * Create the lock table described by LockConflicts
  */
-LOCKMETHOD
+void
 InitLockTable(int maxBackends)
 {
-	int			lockmethod;
+	LOCKMETHODID LongTermTableId;
+
+	/* there's no zero-th table */
+	NumLockMethods = 1;
+
+	/*
+	 * Create the default lock method table
+	 */
 
 	/* number of lock modes is lengthof()-1 because of dummy zero */
-	lockmethod = LockMethodTableInit("LockTable",
-									 LockConflicts,
-									 lengthof(LockConflicts) - 1,
-									 maxBackends);
-	LockTableId = lockmethod;
-
-	if (!(LockTableId))
+	LockTableId = LockMethodTableInit("LockTable",
+									  LockConflicts,
+									  lengthof(LockConflicts) - 1,
+									  maxBackends);
+	if (!LockMethodIsValid(LockTableId))
 		elog(ERROR, "could not initialize lock table");
+	Assert(LockTableId == DEFAULT_LOCKMETHOD);
 
 #ifdef USER_LOCKS
 
@@ -92,11 +99,10 @@ InitLockTable(int maxBackends)
 	 * Allocate another tableId for long-term locks
 	 */
 	LongTermTableId = LockMethodTableRename(LockTableId);
-	if (!(LongTermTableId))
+	if (!LockMethodIsValid(LongTermTableId))
 		elog(ERROR, "could not rename long-term lock table");
+	Assert(LongTermTableId == USER_LOCKMETHOD);
 #endif
-
-	return LockTableId;
 }
 
 /*
@@ -132,7 +138,7 @@ LockRelation(Relation relation, LOCKMODE lockmode)
 	tag.dbId = relation->rd_lockInfo.lockRelId.dbId;
 	tag.objId.blkno = InvalidBlockNumber;
 
-	if (!LockAcquire(LockTableId, &tag, GetCurrentTransactionId(),
+	if (!LockAcquire(LockTableId, &tag, GetTopTransactionId(),
 					 lockmode, false))
 		elog(ERROR, "LockAcquire failed");
 
@@ -166,7 +172,7 @@ ConditionalLockRelation(Relation relation, LOCKMODE lockmode)
 	tag.dbId = relation->rd_lockInfo.lockRelId.dbId;
 	tag.objId.blkno = InvalidBlockNumber;
 
-	if (!LockAcquire(LockTableId, &tag, GetCurrentTransactionId(),
+	if (!LockAcquire(LockTableId, &tag, GetTopTransactionId(),
 					 lockmode, true))
 		return false;
 
@@ -196,7 +202,7 @@ UnlockRelation(Relation relation, LOCKMODE lockmode)
 	tag.dbId = relation->rd_lockInfo.lockRelId.dbId;
 	tag.objId.blkno = InvalidBlockNumber;
 
-	LockRelease(LockTableId, &tag, GetCurrentTransactionId(), lockmode);
+	LockRelease(LockTableId, &tag, GetTopTransactionId(), lockmode);
 }
 
 /*
@@ -259,7 +265,7 @@ LockPage(Relation relation, BlockNumber blkno, LOCKMODE lockmode)
 	tag.dbId = relation->rd_lockInfo.lockRelId.dbId;
 	tag.objId.blkno = blkno;
 
-	if (!LockAcquire(LockTableId, &tag, GetCurrentTransactionId(),
+	if (!LockAcquire(LockTableId, &tag, GetTopTransactionId(),
 					 lockmode, false))
 		elog(ERROR, "LockAcquire failed");
 }
@@ -280,7 +286,7 @@ ConditionalLockPage(Relation relation, BlockNumber blkno, LOCKMODE lockmode)
 	tag.dbId = relation->rd_lockInfo.lockRelId.dbId;
 	tag.objId.blkno = blkno;
 
-	return LockAcquire(LockTableId, &tag, GetCurrentTransactionId(),
+	return LockAcquire(LockTableId, &tag, GetTopTransactionId(),
 					   lockmode, true);
 }
 
@@ -297,7 +303,7 @@ UnlockPage(Relation relation, BlockNumber blkno, LOCKMODE lockmode)
 	tag.dbId = relation->rd_lockInfo.lockRelId.dbId;
 	tag.objId.blkno = blkno;
 
-	LockRelease(LockTableId, &tag, GetCurrentTransactionId(), lockmode);
+	LockRelease(LockTableId, &tag, GetTopTransactionId(), lockmode);
 }
 
 /*
@@ -306,9 +312,6 @@ UnlockPage(Relation relation, BlockNumber blkno, LOCKMODE lockmode)
  * Insert a lock showing that the given transaction ID is running ---
  * this is done during xact startup.  The lock can then be used to wait
  * for the transaction to finish.
- *
- * We need no corresponding unlock function, since the lock will always
- * be released implicitly at transaction commit/abort, never any other way.
  */
 void
 XactLockTableInsert(TransactionId xid)
@@ -320,34 +323,68 @@ XactLockTableInsert(TransactionId xid)
 	tag.dbId = InvalidOid;		/* xids are globally unique */
 	tag.objId.xid = xid;
 
-	if (!LockAcquire(LockTableId, &tag, xid,
+	if (!LockAcquire(LockTableId, &tag, GetTopTransactionId(),
 					 ExclusiveLock, false))
 		elog(ERROR, "LockAcquire failed");
 }
 
+/*
+ *		XactLockTableDelete
+ *
+ * Delete the lock showing that the given transaction ID is running.
+ * (This is never used for main transaction IDs; those locks are only
+ * released implicitly at transaction end.  But we do use it for subtrans
+ * IDs.)
+ */
+void
+XactLockTableDelete(TransactionId xid)
+{
+	LOCKTAG		tag;
+
+	MemSet(&tag, 0, sizeof(tag));
+	tag.relId = XactLockTableId;
+	tag.dbId = InvalidOid;		/* xids are globally unique */
+	tag.objId.xid = xid;
+
+	LockRelease(LockTableId, &tag, GetTopTransactionId(), ExclusiveLock);
+}
+
 /*
  *		XactLockTableWait
  *
  * Wait for the specified transaction to commit or abort.
+ *
+ * Note that this does the right thing for subtransactions: if we wait on a
+ * subtransaction, we will exit as soon as it aborts or its top parent commits.
+ * It takes some extra work to ensure this, because to save on shared memory
+ * the XID lock of a subtransaction is released when it ends, whether
+ * successfully or unsuccessfully.  So we have to check if it's "still running"
+ * and if so wait for its parent.
  */
 void
 XactLockTableWait(TransactionId xid)
 {
 	LOCKTAG		tag;
-	TransactionId myxid = GetCurrentTransactionId();
+	TransactionId myxid = GetTopTransactionId();
 
-	Assert(!TransactionIdEquals(xid, myxid));
+	for (;;)
+	{
+		Assert(TransactionIdIsValid(xid));
+		Assert(!TransactionIdEquals(xid, myxid));
 
-	MemSet(&tag, 0, sizeof(tag));
-	tag.relId = XactLockTableId;
-	tag.dbId = InvalidOid;
-	tag.objId.xid = xid;
+		MemSet(&tag, 0, sizeof(tag));
+		tag.relId = XactLockTableId;
+		tag.dbId = InvalidOid;
+		tag.objId.xid = xid;
 
-	if (!LockAcquire(LockTableId, &tag, myxid,
-					 ShareLock, false))
-		elog(ERROR, "LockAcquire failed");
+		if (!LockAcquire(LockTableId, &tag, myxid, ShareLock, false))
+			elog(ERROR, "LockAcquire failed");
+		LockRelease(LockTableId, &tag, myxid, ShareLock);
 
-	LockRelease(LockTableId, &tag, myxid, ShareLock);
+		if (!TransactionIdIsInProgress(xid))
+			break;
+		xid = SubTransGetParent(xid);
+	}
 
 	/*
 	 * Transaction was committed/aborted/crashed - we have to update
