@@ -3,7 +3,7 @@
  *			  procedural language
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/pl/plpgsql/src/plpgsql.h,v 1.42 2003/09/28 23:37:45 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/pl/plpgsql/src/plpgsql.h,v 1.56.4.1 2008/10/09 16:35:33 tgl Exp $
  *
  *	  This software is copyrighted by Jan Wieck - Hamburg.
  *
@@ -58,8 +58,7 @@ enum
 	PLPGSQL_NSTYPE_LABEL,
 	PLPGSQL_NSTYPE_VAR,
 	PLPGSQL_NSTYPE_ROW,
-	PLPGSQL_NSTYPE_REC,
-	PLPGSQL_NSTYPE_RECFIELD
+	PLPGSQL_NSTYPE_REC
 };
 
 /* ----------
@@ -77,6 +76,18 @@ enum
 	PLPGSQL_DTYPE_TRIGARG
 };
 
+/* ----------
+ * Variants distinguished in PLpgSQL_type structs
+ * ----------
+ */
+enum
+{
+	PLPGSQL_TTYPE_SCALAR,		/* scalar types and domains */
+	PLPGSQL_TTYPE_ROW,			/* composite types */
+	PLPGSQL_TTYPE_REC,			/* RECORD pseudotype */
+	PLPGSQL_TTYPE_PSEUDO		/* other pseudotypes */
+};
+
 /* ----------
  * Execution tree node types
  * ----------
@@ -138,20 +149,21 @@ typedef struct
 	int			alloc;
 	int			used;
 	char	   *value;
-}	PLpgSQL_dstring;
+} PLpgSQL_dstring;
 
 
 typedef struct
-{								/* Postgres data type		*/
-	char	   *typname;
+{								/* Postgres data type */
+	char	   *typname;		/* (simple) name of the type */
 	Oid			typoid;			/* OID of the data type */
+	int			ttype;			/* PLPGSQL_TTYPE_ code */
 	int16		typlen;			/* stuff copied from its pg_type entry */
 	bool		typbyval;
 	Oid			typrelid;
-	Oid			typelem;
+	Oid			typioparam;
 	FmgrInfo	typinput;		/* lookup info for typinput function */
 	int32		atttypmod;		/* typmod (taken from someplace else) */
-}	PLpgSQL_type;
+} PLpgSQL_type;
 
 
 /*
@@ -163,8 +175,19 @@ typedef struct
 {								/* Generic datum array item		*/
 	int			dtype;
 	int			dno;
-}	PLpgSQL_datum;
+} PLpgSQL_datum;
 
+/*
+ * The variants PLpgSQL_var, PLpgSQL_row, and PLpgSQL_rec share these
+ * fields
+ */
+typedef struct
+{								/* Scalar or composite variable */
+	int			dtype;
+	int			dno;
+	char	   *refname;
+	int			lineno;
+} PLpgSQL_variable;
 
 typedef struct PLpgSQL_expr
 {								/* SQL Query to plan and execute	*/
@@ -174,7 +197,7 @@ typedef struct PLpgSQL_expr
 	void	   *plan;
 	Oid		   *plan_argtypes;
 	/* fields for "simple expression" fast-path execution: */
-	Expr	   *expr_simple_expr;	/* NULL means not a simple expr */
+	Expr	   *expr_simple_expr;		/* NULL means not a simple expr */
 	Oid			expr_simple_type;
 	/* if expr is simple AND in use in current xact, these fields are set: */
 	ExprState  *expr_simple_state;
@@ -182,11 +205,11 @@ typedef struct PLpgSQL_expr
 	/* params to pass to expr */
 	int			nparams;
 	int			params[1];		/* VARIABLE SIZE ARRAY ... must be last */
-}	PLpgSQL_expr;
+} PLpgSQL_expr;
 
 
 typedef struct
-{								/* Local variable			*/
+{								/* Scalar variable */
 	int			dtype;
 	int			varno;
 	char	   *refname;
@@ -202,15 +225,16 @@ typedef struct
 	Datum		value;
 	bool		isnull;
 	bool		freeval;
-}	PLpgSQL_var;
+} PLpgSQL_var;
 
 
 typedef struct
-{								/* Rowtype				*/
+{								/* Row variable */
 	int			dtype;
 	int			rowno;
 	char	   *refname;
 	int			lineno;
+
 	TupleDesc	rowtupdesc;
 
 	/*
@@ -223,11 +247,11 @@ typedef struct
 	int			nfields;
 	char	  **fieldnames;
 	int		   *varnos;
-}	PLpgSQL_row;
+} PLpgSQL_row;
 
 
 typedef struct
-{								/* Record of non-fixed structure */
+{								/* Record variable (non-fixed structure) */
 	int			dtype;
 	int			recno;
 	char	   *refname;
@@ -237,7 +261,7 @@ typedef struct
 	TupleDesc	tupdesc;
 	bool		freetup;
 	bool		freetupdesc;
-}	PLpgSQL_rec;
+} PLpgSQL_rec;
 
 
 typedef struct
@@ -246,7 +270,7 @@ typedef struct
 	int			rfno;
 	char	   *fieldname;
 	int			recparentno;	/* dno of parent record */
-}	PLpgSQL_recfield;
+} PLpgSQL_recfield;
 
 
 typedef struct
@@ -255,7 +279,7 @@ typedef struct
 	int			dno;
 	PLpgSQL_expr *subscript;
 	int			arrayparentno;	/* dno of parent array variable */
-}	PLpgSQL_arrayelem;
+} PLpgSQL_arrayelem;
 
 
 typedef struct
@@ -263,7 +287,7 @@ typedef struct
 	int			dtype;
 	int			dno;
 	PLpgSQL_expr *argnum;
-}	PLpgSQL_trigarg;
+} PLpgSQL_trigarg;
 
 
 typedef struct
@@ -271,7 +295,7 @@ typedef struct
 	int			itemtype;
 	int			itemno;
 	char		name[1];
-}	PLpgSQL_nsitem;
+} PLpgSQL_nsitem;
 
 
 typedef struct PLpgSQL_ns
@@ -280,22 +304,46 @@ typedef struct PLpgSQL_ns
 	int			items_used;
 	PLpgSQL_nsitem **items;
 	struct PLpgSQL_ns *upper;
-}	PLpgSQL_ns;
-
-
-typedef struct
-{								/* List of execution nodes		*/
-	int			stmts_alloc;
-	int			stmts_used;
-	struct PLpgSQL_stmt **stmts;
-}	PLpgSQL_stmts;
+} PLpgSQL_ns;
 
 
 typedef struct
 {								/* Generic execution node		*/
 	int			cmd_type;
 	int			lineno;
-}	PLpgSQL_stmt;
+} PLpgSQL_stmt;
+
+
+typedef struct
+{								/* List of execution nodes		*/
+	int			stmts_alloc;	/* XXX this oughta just be a List ... */
+	int			stmts_used;
+	PLpgSQL_stmt **stmts;
+} PLpgSQL_stmts;
+
+
+typedef struct PLpgSQL_condition
+{								/* One EXCEPTION condition name */
+	int			sqlerrstate;	/* SQLSTATE code */
+	char	   *condname;		/* condition name (for debugging) */
+	struct PLpgSQL_condition *next;
+} PLpgSQL_condition;
+
+typedef struct
+{								/* One EXCEPTION ... WHEN clause */
+	int			lineno;
+	PLpgSQL_condition *conditions;
+	PLpgSQL_stmts *action;
+} PLpgSQL_exception;
+
+
+typedef struct
+{								/* List of WHEN clauses			*/
+	int			exceptions_alloc;		/* XXX this oughta just be a List
+										 * ... */
+	int			exceptions_used;
+	PLpgSQL_exception **exceptions;
+} PLpgSQL_exceptions;
 
 
 typedef struct
@@ -304,9 +352,10 @@ typedef struct
 	int			lineno;
 	char	   *label;
 	PLpgSQL_stmts *body;
+	PLpgSQL_exceptions *exceptions;
 	int			n_initvars;
 	int		   *initvarnos;
-}	PLpgSQL_stmt_block;
+} PLpgSQL_stmt_block;
 
 
 typedef struct
@@ -315,20 +364,20 @@ typedef struct
 	int			lineno;
 	int			varno;
 	PLpgSQL_expr *expr;
-}	PLpgSQL_stmt_assign;
+} PLpgSQL_stmt_assign;
 
 typedef struct
 {								/* PERFORM statement		*/
 	int			cmd_type;
 	int			lineno;
 	PLpgSQL_expr *expr;
-}	PLpgSQL_stmt_perform;
+} PLpgSQL_stmt_perform;
 
 typedef struct
 {								/* Get Diagnostics item		*/
 	int			item;			/* id for diagnostic value desired */
 	int			target;			/* where to assign it */
-}	PLpgSQL_diag_item;
+} PLpgSQL_diag_item;
 
 typedef struct
 {								/* Get Diagnostics statement		*/
@@ -336,7 +385,7 @@ typedef struct
 	int			lineno;
 	int			ndtitems;
 	PLpgSQL_diag_item *dtitems;
-}	PLpgSQL_stmt_getdiag;
+} PLpgSQL_stmt_getdiag;
 
 
 typedef struct
@@ -346,7 +395,7 @@ typedef struct
 	PLpgSQL_expr *cond;
 	PLpgSQL_stmts *true_body;
 	PLpgSQL_stmts *false_body;
-}	PLpgSQL_stmt_if;
+} PLpgSQL_stmt_if;
 
 
 typedef struct
@@ -355,7 +404,7 @@ typedef struct
 	int			lineno;
 	char	   *label;
 	PLpgSQL_stmts *body;
-}	PLpgSQL_stmt_loop;
+} PLpgSQL_stmt_loop;
 
 
 typedef struct
@@ -365,7 +414,7 @@ typedef struct
 	char	   *label;
 	PLpgSQL_expr *cond;
 	PLpgSQL_stmts *body;
-}	PLpgSQL_stmt_while;
+} PLpgSQL_stmt_while;
 
 
 typedef struct
@@ -378,7 +427,7 @@ typedef struct
 	PLpgSQL_expr *upper;
 	int			reverse;
 	PLpgSQL_stmts *body;
-}	PLpgSQL_stmt_fori;
+} PLpgSQL_stmt_fori;
 
 
 typedef struct
@@ -390,7 +439,7 @@ typedef struct
 	PLpgSQL_row *row;
 	PLpgSQL_expr *query;
 	PLpgSQL_stmts *body;
-}	PLpgSQL_stmt_fors;
+} PLpgSQL_stmt_fors;
 
 
 typedef struct
@@ -402,7 +451,7 @@ typedef struct
 	PLpgSQL_row *row;
 	PLpgSQL_expr *query;
 	PLpgSQL_stmts *body;
-}	PLpgSQL_stmt_dynfors;
+} PLpgSQL_stmt_dynfors;
 
 
 typedef struct
@@ -412,7 +461,7 @@ typedef struct
 	PLpgSQL_rec *rec;
 	PLpgSQL_row *row;
 	PLpgSQL_expr *query;
-}	PLpgSQL_stmt_select;
+} PLpgSQL_stmt_select;
 
 
 typedef struct
@@ -424,7 +473,7 @@ typedef struct
 	PLpgSQL_expr *argquery;
 	PLpgSQL_expr *query;
 	PLpgSQL_expr *dynquery;
-}	PLpgSQL_stmt_open;
+} PLpgSQL_stmt_open;
 
 
 typedef struct
@@ -434,7 +483,7 @@ typedef struct
 	PLpgSQL_rec *rec;
 	PLpgSQL_row *row;
 	int			curvar;
-}	PLpgSQL_stmt_fetch;
+} PLpgSQL_stmt_fetch;
 
 
 typedef struct
@@ -442,7 +491,7 @@ typedef struct
 	int			cmd_type;
 	int			lineno;
 	int			curvar;
-}	PLpgSQL_stmt_close;
+} PLpgSQL_stmt_close;
 
 
 typedef struct
@@ -451,7 +500,7 @@ typedef struct
 	int			lineno;
 	char	   *label;
 	PLpgSQL_expr *cond;
-}	PLpgSQL_stmt_exit;
+} PLpgSQL_stmt_exit;
 
 
 typedef struct
@@ -461,7 +510,7 @@ typedef struct
 	PLpgSQL_expr *expr;
 	int			retrecno;
 	int			retrowno;
-}	PLpgSQL_stmt_return;
+} PLpgSQL_stmt_return;
 
 typedef struct
 {								/* RETURN NEXT statement */
@@ -470,7 +519,7 @@ typedef struct
 	PLpgSQL_rec *rec;
 	PLpgSQL_row *row;
 	PLpgSQL_expr *expr;
-}	PLpgSQL_stmt_return_next;
+} PLpgSQL_stmt_return_next;
 
 typedef struct
 {								/* RAISE statement			*/
@@ -480,7 +529,7 @@ typedef struct
 	char	   *message;
 	int			nparams;
 	int		   *params;
-}	PLpgSQL_stmt_raise;
+} PLpgSQL_stmt_raise;
 
 
 typedef struct
@@ -488,7 +537,7 @@ typedef struct
 	int			cmd_type;
 	int			lineno;
 	PLpgSQL_expr *sqlstmt;
-}	PLpgSQL_stmt_execsql;
+} PLpgSQL_stmt_execsql;
 
 
 typedef struct
@@ -496,18 +545,22 @@ typedef struct
 	int			cmd_type;
 	int			lineno;
 	PLpgSQL_expr *query;
-}	PLpgSQL_stmt_dynexecute;
+} PLpgSQL_stmt_dynexecute;
 
 
 typedef struct PLpgSQL_func_hashkey
 {								/* Hash lookup key for functions */
 	Oid			funcOid;
 
+	bool		isTrigger;		/* true if called as a trigger */
+
+	/* be careful that pad bytes in this struct get zeroed! */
+
 	/*
-	 * For a trigger function, the OID of the relation triggered on is part
-	 * of the hashkey --- we want to compile the trigger separately for each
-	 * relation it is used with, in case the rowtype is different.  Zero if
-	 * not called as a trigger.
+	 * For a trigger function, the OID of the relation triggered on is
+	 * part of the hashkey --- we want to compile the trigger separately
+	 * for each relation it is used with, in case the rowtype is
+	 * different.  Zero if not called as a trigger.
 	 */
 	Oid			trigrelOid;
 
@@ -517,7 +570,7 @@ typedef struct PLpgSQL_func_hashkey
 	 * zeroed!
 	 */
 	Oid			argtypes[FUNC_MAX_ARGS];
-}	PLpgSQL_func_hashkey;
+} PLpgSQL_func_hashkey;
 
 
 typedef struct PLpgSQL_function
@@ -533,9 +586,10 @@ typedef struct PLpgSQL_function
 	int			fn_rettyplen;
 	bool		fn_retbyval;
 	FmgrInfo	fn_retinput;
-	Oid			fn_rettypelem;
+	Oid			fn_rettypioparam;
 	bool		fn_retistuple;
 	bool		fn_retset;
+	bool		fn_readonly;
 
 	int			fn_nargs;
 	int			fn_argvarnos[FUNC_MAX_ARGS];
@@ -553,7 +607,7 @@ typedef struct PLpgSQL_function
 	int			ndatums;
 	PLpgSQL_datum **datums;
 	PLpgSQL_stmt_block *action;
-}	PLpgSQL_function;
+} PLpgSQL_function;
 
 
 typedef struct
@@ -566,6 +620,8 @@ typedef struct
 	bool		retistuple;
 	bool		retisset;
 
+	bool		readonly_func;
+
 	TupleDesc	rettupdesc;
 	char	   *exitlabel;
 
@@ -590,7 +646,7 @@ typedef struct
 	PLpgSQL_function *err_func; /* current func */
 	PLpgSQL_stmt *err_stmt;		/* current stmt */
 	const char *err_text;		/* additional state info */
-}	PLpgSQL_execstate;
+} PLpgSQL_execstate;
 
 
 /**********************************************************************
@@ -620,7 +676,8 @@ extern PLpgSQL_function *plpgsql_curr_compile;
  * Functions in pl_comp.c
  * ----------
  */
-extern PLpgSQL_function *plpgsql_compile(FunctionCallInfo fcinfo);
+extern PLpgSQL_function *plpgsql_compile(FunctionCallInfo fcinfo,
+				bool forValidator);
 extern int	plpgsql_parse_word(char *word);
 extern int	plpgsql_parse_dblword(char *word);
 extern int	plpgsql_parse_tripword(char *word);
@@ -629,11 +686,14 @@ extern int	plpgsql_parse_dblwordtype(char *word);
 extern int	plpgsql_parse_tripwordtype(char *word);
 extern int	plpgsql_parse_wordrowtype(char *word);
 extern int	plpgsql_parse_dblwordrowtype(char *word);
-extern PLpgSQL_type *plpgsql_parse_datatype(char *string);
-extern PLpgSQL_row *plpgsql_build_rowtype(Oid classOid);
-extern void plpgsql_adddatum(PLpgSQL_datum * new);
+extern PLpgSQL_type *plpgsql_parse_datatype(const char *string);
+extern PLpgSQL_type *plpgsql_build_datatype(Oid typeOid, int32 typmod);
+extern PLpgSQL_variable *plpgsql_build_variable(char *refname, int lineno,
+					   PLpgSQL_type *dtype,
+					   bool add2namespace);
+extern PLpgSQL_condition *plpgsql_parse_err_condition(char *condname);
+extern void plpgsql_adddatum(PLpgSQL_datum *new);
 extern int	plpgsql_add_initdatums(int **varnos);
-extern void plpgsql_yyerror(const char *s);
 extern void plpgsql_HashTableInit(void);
 
 /* ----------
@@ -642,25 +702,26 @@ extern void plpgsql_HashTableInit(void);
  */
 extern void plpgsql_init(void);
 extern Datum plpgsql_call_handler(PG_FUNCTION_ARGS);
+extern Datum plpgsql_validator(PG_FUNCTION_ARGS);
 
 /* ----------
  * Functions in pl_exec.c
  * ----------
  */
-extern Datum plpgsql_exec_function(PLpgSQL_function * func,
+extern Datum plpgsql_exec_function(PLpgSQL_function *func,
 					  FunctionCallInfo fcinfo);
-extern HeapTuple plpgsql_exec_trigger(PLpgSQL_function * func,
+extern HeapTuple plpgsql_exec_trigger(PLpgSQL_function *func,
 					 TriggerData *trigdata);
-extern void plpgsql_eoxact(bool isCommit, void *arg);
+extern void plpgsql_xact_cb(XactEvent event, void *arg);
 
 /* ----------
  * Functions for the dynamic string handling in pl_funcs.c
  * ----------
  */
-extern void plpgsql_dstring_init(PLpgSQL_dstring * ds);
-extern void plpgsql_dstring_free(PLpgSQL_dstring * ds);
-extern void plpgsql_dstring_append(PLpgSQL_dstring * ds, char *str);
-extern char *plpgsql_dstring_get(PLpgSQL_dstring * ds);
+extern void plpgsql_dstring_init(PLpgSQL_dstring *ds);
+extern void plpgsql_dstring_free(PLpgSQL_dstring *ds);
+extern void plpgsql_dstring_append(PLpgSQL_dstring *ds, const char *str);
+extern char *plpgsql_dstring_get(PLpgSQL_dstring *ds);
 
 /* ----------
  * Functions for the namestack handling in pl_funcs.c
@@ -679,8 +740,8 @@ extern void plpgsql_ns_rename(char *oldname, char *newname);
  * ----------
  */
 extern void plpgsql_convert_ident(const char *s, char **output, int numidents);
-extern const char *plpgsql_stmt_typename(PLpgSQL_stmt * stmt);
-extern void plpgsql_dumptree(PLpgSQL_function * func);
+extern const char *plpgsql_stmt_typename(PLpgSQL_stmt *stmt);
+extern void plpgsql_dumptree(PLpgSQL_function *func);
 
 /* ----------
  * Externs in gram.y and scan.l
@@ -691,8 +752,10 @@ extern int	plpgsql_yyparse(void);
 extern int	plpgsql_base_yylex(void);
 extern int	plpgsql_yylex(void);
 extern void plpgsql_push_back_token(int token);
+extern void plpgsql_yyerror(const char *message);
 extern int	plpgsql_scanner_lineno(void);
 extern void plpgsql_scanner_init(const char *str, int functype);
 extern void plpgsql_scanner_finish(void);
+extern char *plpgsql_get_string_value(void);
 
 #endif   /* PLPGSQL_H */
